#!/usr/bin/env python
import seaborn as sns
import numpy as np
from matplotlib import pyplot as plt

sns.set_style('white')
plt.rcParams['svg.fonttype'] = 'none'


def plot_contrast_matrix(contrast_matrix, ornt='vertical', ax=None):
    """ Plot correlation matrix

    Parameters
    ----------
    mat : DataFrame
        Design matrix with columns consisting of explanatory variables followed
        by confounds
    n_evs : int
        Number of explanatory variables to separate from confounds
    partial : {'upper', 'lower', None}, optional
        Plot matrix as upper triangular (default), lower triangular or full

    Returns
    -------
    ax : Axes
        Axes containing plot
    """

    if ax is None:
        plt.figure()
        ax = plt.gca()

    if ornt == 'horizontal':
        contrast_matrix = contrast_matrix.T

    vmax = np.abs(contrast_matrix.values).max()

    # Use a red/blue (+1/-1) diverging colormap
    cmap = sns.diverging_palette(220, 10, as_cmap=True)

    sns.heatmap(contrast_matrix, vmin=-vmax, vmax=vmax, square=True,
                linewidths=0.5, cmap=cmap,
                cbar_kws={'shrink': 0.5, 'orientation': ornt,
                          'ticks': np.linspace(-vmax, vmax, 5)},
                ax=ax)

    # Variables along top and left
    ax.xaxis.tick_top()
    xtl = ax.get_xticklabels()
    ax.set_xticklabels(xtl, rotation=90)

    return ax



#!/usr/bin/env python
import seaborn as sns
import numpy as np
from matplotlib import pyplot as plt

sns.set_style('white')
plt.rcParams['svg.fonttype'] = 'none'


def plot_corr_matrix(mat, n_evs, partial='upper', ax=None):
    """ Plot correlation matrix

    Parameters
    ----------
    mat : DataFrame
        Design matrix with columns consisting of explanatory variables followed
        by confounds
    n_evs : int
        Number of explanatory variables to separate from confounds
    partial : {'upper', 'lower', None}, optional
        Plot matrix as upper triangular (default), lower triangular or full

    Returns
    -------
    ax : Axes
        Axes containing plot
    """

    if ax is None:
        ax = plt.figure().subplot(1, 1, 1)

    # For a heatmap mask, 0 = show, 1 = hide
    mask = None if partial is None else np.ones_like(mat, dtype=bool)
    if partial == 'upper':
        mask[np.triu_indices_from(mask)] = False
    elif partial == 'lower':
        mask[np.tril_indices_from(mask)] = False
    elif partial is not None:
        raise ValueError("partial must be 'upper' or 'lower' or None")

    # Use a red/blue (+1/-1) diverging colormap
    cmap = sns.diverging_palette(220, 10, as_cmap=True)

    sns.heatmap(mat, vmin=-1, vmax=1, mask=mask, square=True,
                linewidths=0.5, cmap=cmap, cbar_kws={'shrink': 0.5}, ax=ax)

    # Separate EV-EV, EV-Confound, Confound-Confound correlation sections with
    # black line
    if partial == 'upper':
        ax.hlines([n_evs], n_evs, len(mat))
        ax.vlines([n_evs], 0, n_evs)
    elif partial == 'lower':
        ax.hlines([n_evs], 0, n_evs)
        ax.vlines([n_evs], n_evs, len(mat))
    else:
        ax.hlines([n_evs], 0, len(mat))
        ax.vlines([n_evs], 0, len(mat))

    # Label with variable names
    if partial == 'upper':
        # Upper: variables along top, and following the diagonal to the left
        ax.xaxis.tick_top()
        xtl = ax.get_xticklabels()
        ax.set_xticklabels(xtl, rotation=90)
        ax.set_yticklabels([])
        for i, var in enumerate(mat.columns):
            ax.text(i - 0.2, i + 0.6, var, ha="right", va="center")
    elif partial == 'lower':
        # Lower: Variables along left and following the diagonal to the top
        ax.set_xticklabels([])
        for i, var in enumerate(mat.columns):
            ax.text(i + 0.6, i - 0.2, var, ha="center", va="bottom", rotation='90')
    else:
        # Full: Variables along top and left
        ax.xaxis.tick_top()
        xtl = ax.get_xticklabels()
        ax.set_xticklabels(xtl, rotation=90)

    return ax



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Base module variables
"""

__version__ = '0.0.1'
__author__ = 'The CRN developers'
__copyright__ = ('Copyright 2018, Center for Reproducible Neuroscience, '
                 'Stanford University')
__credits__ = ['Christopher J. Markiewicz', 'Chris Gorgolewski',
               'Russell A. Poldrack']
__license__ = '3-clause BSD'
__maintainer__ = 'Christopher J. Markiewicz'
__email__ = 'crn.poldracklab@gmail.com'
__status__ = 'Prototype'
__url__ = 'https://github.com/poldracklab/fitlins'
__packagename__ = 'fitlins'
__description__ = 'Fit Linear Models to fMRI Data'
__longdesc__ = ''

DOWNLOAD_URL = (
    'https://github.com/poldracklab/{name}/archive/{ver}.tar.gz'.format(
        name=__packagename__, ver=__version__))


SETUP_REQUIRES = [
    'setuptools>=27.0',
]

REQUIRES = [
    'nibabel>=2.0',
    'seaborn>=0.7.1',
    'numpy>=1.11',
    'nilearn>=0.4.0',
    'pandas>=0.19',
    'nistats>=0.0.1a',
    'pybids>=0.3',
    'jinja2',
]

LINKS_REQUIRES = [
    'git+https://github.com/grabbles/grabbit.git@'
    'f317224f90f572ec35899a43f19cb82b870ff772#egg=grabbit-0.1.1-dev',
    'git+https://github.com/INCF/pybids.git@'
    'ea6cf9db042f25dfcae907bd535e9dcabcf2bd55#egg=pybids-0.4.3-dev',
    'git+https://github.com/effigies/nistats.git@'
    'd5b63e0ce1ea817bacd9117b9d729b3ca62c9352#egg=nistats-0.0.1b-dev',
]

TESTS_REQUIRES = [
]

EXTRA_REQUIRES = {
    'tests': TESTS_REQUIRES,
    'duecredit': ['duecredit'],
}

# Enable a handle to install all extra dependencies at once
EXTRA_REQUIRES['all'] = [val for _, val in list(EXTRA_REQUIRES.items())]
CLASSIFIERS = [
    'Development Status :: 2 - Pre-Alpha',
    'Intended Audience :: Science/Research',
    'Topic :: Scientific/Engineering',
    'License :: OSI Approved :: BSD License',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
]



from os import path as op
import jinja2
import pkg_resources as pkgr
from bids import grabbids

from ..utils import snake_to_camel

PATH_PATTERNS = [
    '[sub-{subject}/][ses-{session}/][sub-{subject}_][ses-{session}_]'
    'model-{model}.html'
]


def deroot(val, root):
    if isinstance(val, str):
        if val.startswith(root):
            idx = len(root)
            if val[idx] == '/':
                idx += 1
            val = val[idx:]
    elif isinstance(val, list):
        val = [deroot(elem, root) for elem in val]
    elif isinstance(val, dict):
        val = {key: deroot(value, root) for key, value in val.items()}

    return val


def write_report(level, report_dicts, run_context, deriv_dir):
    fl_layout = grabbids.BIDSLayout(
        deriv_dir,
        extensions=['derivatives',
                    pkgr.resource_filename('fitlins', 'data/fitlins.json')])
    fl_layout.path_patterns = PATH_PATTERNS

    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(
            searchpath=pkgr.resource_filename('fitlins', '/')))

    tpl = env.get_template('data/report.tpl')

    for context in report_dicts:
        ents = context['ents'].copy()
        ents['model'] = snake_to_camel(context['model_name'])
        target_file = op.join(deriv_dir, fl_layout.build_path(ents))
        html = tpl.render(deroot({'level': level, **context, **run_context},
                                 op.dirname(target_file)))
        with open(target_file, 'w') as fobj:
            fobj.write(html)



import seaborn as sns
import numpy as np
from matplotlib import pyplot as plt

import os
from os import path as op
from functools import reduce
from scipy import stats as sps
import pandas as pd
import nibabel as nb
from nilearn import plotting as nlp
import nistats as nis
import nistats.reporting
from nistats import design_matrix as dm
from nistats import first_level_model as level1, second_level_model as level2

import pkg_resources as pkgr

from grabbit import merge_layouts
from bids import grabbids
from bids import analysis as ba

from fitlins.utils import dict_intersection, snake_to_camel
from fitlins.viz import plot_and_save, plot_corr_matrix, plot_contrast_matrix

sns.set_style('white')
plt.rcParams['svg.fonttype'] = 'none'
plt.rcParams['image.interpolation'] = 'nearest'

PATH_PATTERNS = (
    '[sub-{subject}/][ses-{session}/][sub-{subject}_][ses-{session}_]'
    'task-{task}_bold[_space-{space}]_contrast-{contrast}_{type<stat>}.nii.gz',
    '[sub-{subject}/][ses-{session}/][sub-{subject}_][ses-{session}_]'
    'task-{task}_bold[_space-{space}]_contrast-{contrast}_{type<ortho>}.png',
    'sub-{subject}/[ses-{session}/]sub-{subject}_[ses-{session}_]'
    'task-{task}_bold_{type<design>}.tsv',
    '[sub-{subject}/][ses-{session}/][sub-{subject}_][ses-{session}_]'
    'task-{task}_bold_{type<corr|contrasts>}.svg',
    )


def expand_contrast_matrix(contrast_matrix, design_matrix):
    """ Ensure contrast matrix has entries for every regressor

    Parameters
    ----------
    contrast_matrix : DataFrame
        Weight matrix with contrasts as columns and regressors as rows
    design_matrix : DataFrame
        GLM design matrix with regressors as columns and TR time as rows

    Returns
    -------
    contrast_matrix : DataFrame
        Updated matrix with row for every column in ``design_matrix``, matching
        column order
    """
    full_mat = pd.DataFrame(index=design_matrix.columns,
                            columns=contrast_matrix.columns, data=0)
    full_mat.loc[contrast_matrix.index,
                 contrast_matrix.columns] = contrast_matrix

    return full_mat


def init(model_fname, bids_dir, preproc_dir):
    orig_layout = grabbids.BIDSLayout(bids_dir)
    prep_layout = grabbids.BIDSLayout(preproc_dir, extensions=['derivatives'])
    analysis = ba.Analysis(model=model_fname,
                           layout=merge_layouts([orig_layout, prep_layout]))
    analysis.setup(**analysis.model['input'])
    analysis.layout.path_patterns[:0] = PATH_PATTERNS
    return analysis


def first_level(analysis, block, space, deriv_dir):
    analyses = []
    for paradigm, _, ents in block.get_design_matrix(block.model['HRF_variables'],
                                                     mode='sparse'):
        preproc_files = analysis.layout.get(type='preproc', space=space, **ents)
        if len(preproc_files) == 0:
            raise ValueError("No PREPROC files found")

        if len(preproc_files) != 1:
            print(preproc_files)
            raise ValueError("Too many potential PREPROC files")

        fname = preproc_files[0].filename

        img = nb.load(fname)
        TR = img.header.get_zooms()[3]
        vols = img.shape[3]

        # Get dense portion of design matrix once TR is known
        _, confounds, _ = block.get_design_matrix(mode='dense',
                                                  sampling_rate=1/TR, **ents)[0]
        names = [col for col in confounds.columns
                 if col.startswith('NonSteadyStateOutlier') or
                 col in block.model['variables']]

        mat = dm.make_design_matrix(
            frame_times=np.arange(vols) * TR,
            paradigm=paradigm.rename(columns={'condition': 'trial_type',
                                              'amplitude': 'modulation'}),
            add_regs=confounds[names].fillna(0),
            add_reg_names=names,
            drift_model=None if 'Cosine00' in names else 'cosine',
            )

        preproc_ents = analysis.layout.parse_entities(fname)

        dm_ents = {k: v for k, v in preproc_ents.items()
                   if k in ('subject', 'session', 'task')}

        dm_ents['type'] = 'design'
        design_fname = op.join(deriv_dir,
                               analysis.layout.build_path(dm_ents, strict=True))
        os.makedirs(op.dirname(design_fname), exist_ok=True)
        mat.to_csv(design_fname, sep='\t')
        plt.set_cmap('viridis')
        plot_and_save(design_fname.replace('.tsv', '.svg'),
                      nis.reporting.plot_design_matrix, mat)

        corr_ents = dm_ents.copy()
        corr_ents['type'] = 'corr'
        corr_fname = op.join(deriv_dir,
                             analysis.layout.build_path(corr_ents, strict=True))
        plot_and_save(corr_fname, plot_corr_matrix,
                      mat.drop(columns=['constant']).corr(),
                      len(block.model['HRF_variables']))

        job_desc = {
            'ents': ents,
            'subject_id': ents['subject'],
            'dataset': analysis.layout.root,
            'model_name': analysis.model['name'],
            'design_matrix_svg': design_fname.replace('.tsv', '.svg'),
            'correlation_matrix_svg': corr_fname,
            }

        cnames = [contrast['name'] for contrast in block.contrasts] + block.model['HRF_variables']
        contrast_matrix = []
        if cnames:
            contrasts_ents = corr_ents.copy()
            contrasts_ents['type'] = 'contrasts'
            contrasts_fname = op.join(
                deriv_dir,
                analysis.layout.build_path(contrasts_ents, strict=True))

            contrast_matrix = expand_contrast_matrix(
                block.get_contrasts(cnames, **ents)[0][0], mat)
            plot_and_save(contrasts_fname, plot_contrast_matrix,
                          contrast_matrix.drop(['constant'], 'index'),
                          ornt='horizontal')

            job_desc['contrasts_svg'] = contrasts_fname

        brainmask = analysis.layout.get(type='brainmask', space=space,
                                        **ents)[0]
        fmri_glm = None

        for contrast in contrast_matrix:
            stat_ents = preproc_ents.copy()
            stat_ents.pop('modality', None)
            stat_ents.update({'contrast': snake_to_camel(contrast),
                              'type': 'stat'})
            stat_fname = op.join(deriv_dir,
                                 analysis.layout.build_path(stat_ents,
                                                            strict=True))

            ortho_ents = stat_ents.copy()
            ortho_ents['type'] = 'ortho'
            ortho_fname = op.join(deriv_dir,
                                  analysis.layout.build_path(ortho_ents,
                                                             strict=True))

            desc = {'name': contrast, 'image_file': ortho_fname}
            if contrast not in block.model['HRF_variables']:
                job_desc.setdefault('contrasts', []).append(desc)
            else:
                job_desc.setdefault('estimates', []).append(desc)

            if op.exists(stat_fname):
                continue

            if fmri_glm is None:
                fmri_glm = level1.FirstLevelModel(mask=brainmask.filename)
                fmri_glm.fit(fname, design_matrices=mat)

            stat_types = [c['type'] for c in block.contrasts if c['name'] == contrast]
            stat_type = stat_types[0] if stat_types else 'T'
            stat = fmri_glm.compute_contrast(contrast_matrix[contrast].values,
                                             {'T': 't', 'F': 'F'}[stat_type])
            stat.to_filename(stat_fname)

            nlp.plot_glass_brain(stat, colorbar=True, plot_abs=False,
                                 display_mode='lyrz', axes=None,
                                 output_file=ortho_fname)

        analyses.append(job_desc)

    return analyses


def second_level(analysis, block, space, deriv_dir):
    fl_layout = grabbids.BIDSLayout(
        deriv_dir,
        extensions=['derivatives',
                    pkgr.resource_filename('fitlins', 'data/fitlins.json')])
    fl_layout.path_patterns[:0] = PATH_PATTERNS

    analyses = []

    # pybids likes to give us a lot of extraneous columns
    cnames = [contrast['name'] for contrast in block.contrasts]
    fmri_glm = level2.SecondLevelModel()
    for contrasts, idx, ents in block.get_contrasts(names=cnames):
        if contrasts.empty:
            continue

        data = []
        for in_name, sub_ents in zip(contrasts.index, idx.to_dict(orient='record')):
            # The underlying contrast name might have been added to by a transform
            for option in [in_name] + in_name.split('.'):
                files = fl_layout.get(contrast=snake_to_camel(option),
                                      type='stat', space=space, **sub_ents)
                if files:
                    data.append(files[0].filename)
                    break
            else:
                raise ValueError("Unknown input: {}".format(in_name))

        out_ents = reduce(dict_intersection,
                          map(fl_layout.parse_entities, data))

        contrasts_ents = out_ents.copy()
        contrasts_ents['type'] = 'contrasts'
        contrasts_ents.pop('contrast', None)
        contrasts_ents.pop('space', None)
        contrasts_fname = op.join(
            deriv_dir,
            fl_layout.build_path(contrasts_ents, strict=True))

        plot_and_save(contrasts_fname, plot_contrast_matrix, contrasts,
                      ornt='horizontal')

        job_desc = {
            'ents': out_ents,
            'subject_id': ents.get('subject'),
            'dataset': analysis.layout.root,
            'model_name': analysis.model['name'],
            'contrasts_svg': contrasts_fname,
            }

        for contrast in contrasts:
            out_ents['contrast'] = snake_to_camel(contrast)

            stat_fname = op.join(deriv_dir,
                                 fl_layout.build_path(out_ents, strict=True))

            ortho_ents = out_ents.copy()
            ortho_ents['type'] = 'ortho'
            ortho_fname = op.join(deriv_dir,
                                  analysis.layout.build_path(ortho_ents,
                                                             strict=True))

            desc = {'name': contrast, 'image_file': ortho_fname}
            job_desc.setdefault('contrasts', []).append(desc)

            if op.exists(stat_fname):
                continue

            cols = {'intercept': np.ones(len(data))}
            cname = 'intercept'
            if not np.allclose(contrasts[contrast], 1):
                cname = contrast
                cols[contrast] = contrasts[contrast]

            paradigm = pd.DataFrame(cols)

            fmri_glm.fit(data, design_matrix=paradigm)
            stat_type = [c['type'] for c in block.contrasts if c['name'] == contrast][0]
            stat = fmri_glm.compute_contrast(
                cname,
                second_level_stat_type={'T': 't', 'F': 'F'}[stat_type],
                )
            data = stat.get_data()
            masked_vals = data[data != 0]
            if np.isnan(masked_vals).all():
                raise ValueError("nistats was unable to perform this contrast")
            stat.to_filename(stat_fname)

            nlp.plot_glass_brain(stat, colorbar=True,
                                 threshold=sps.norm.isf(0.001), plot_abs=False,
                                 display_mode='lyrz', output_file=ortho_fname)

        analyses.append(job_desc)

    return analyses



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
fMRI model-fitting
==================
"""

import sys
import os
import os.path as op
import time
import logging
import warnings
from argparse import ArgumentParser
from argparse import RawTextHelpFormatter

from .. import info
from ..utils import bids
from ..base import init, first_level, second_level
from ..viz.reports import write_report

logging.addLevelName(25, 'INFO')  # Add a new level between INFO and WARNING
logger = logging.getLogger('cli')
logger.setLevel(25)

INIT_MSG = """
Running FitLins version {version}:
  * Participant list: {subject_list}.
""".format


def _warn_redirect(message, category, filename, lineno, file=None, line=None):
    logger.warning('Captured warning (%s): %s', category, message)


def default_path(arg, base_dir, default_name):
    """Generate absolute path from absolute, relative, or missing path

    Missing paths are given a default value, which may be absolute or relative,
    and relative paths are resolved relative to a base directory.
    """
    path = default_name if arg is None else arg
    if not op.isabs(path):
        path = op.abspath(op.join(base_dir, path))
    return path


def get_parser():
    """Build parser object"""
    verstr = 'fitlins v{}'.format(info.__version__)

    parser = ArgumentParser(description='FitLins: Workflows for Fitting Linear models to fMRI',
                            formatter_class=RawTextHelpFormatter)

    # Arguments as specified by BIDS-Apps
    # required, positional arguments
    # IMPORTANT: they must go directly with the parser object
    parser.add_argument('bids_dir', action='store',
                        help='the root folder of a BIDS valid dataset (sub-XXXXX folders should '
                             'be found at the top level in this folder).')
    parser.add_argument('output_dir', action='store',
                        help='the output path for the outcomes of preprocessing and visual '
                             'reports')
    parser.add_argument('analysis_level', choices=['run', 'session', 'participant', 'dataset'],
                        help='processing stage to be runa (see BIDS-Apps specification).')

    # optional arguments
    parser.add_argument('-v', '--version', action='version', version=verstr)

    g_bids = parser.add_argument_group('Options for filtering BIDS queries')
    g_bids.add_argument('--participant-label', action='store', nargs='+',
                        help='one or more participant identifiers (the sub- prefix can be '
                             'removed)')
    g_bids.add_argument('-m', '--model', action='store', default='model.json',
                        help='location of BIDS model description (default bids_dir/model.json)')
    g_bids.add_argument('-p', '--preproc-dir', action='store', default='fmriprep',
                        help='location of preprocessed data (default output_dir/fmriprep)')
    g_bids.add_argument('--space', action='store',
                        choices=['MNI152NLin2009cAsym'], default='MNI152NLin2009cAsym',
                        help='registered space of input datasets')

    g_perfm = parser.add_argument_group('Options to handle performance')
    g_perfm.add_argument('--debug', action='store_true', default=False,
                         help='run debug version of workflow')

    return parser


def main(args=None):
    """Entry point"""
    warnings.showwarning = _warn_redirect
    opts = get_parser().parse_args(args)
    if opts.debug:
        logger.setLevel(logging.DEBUG)

    create_workflow(opts)


def create_workflow(opts):
    """Build workflow"""

    # First check that bids_dir looks like a BIDS folder
    bids_dir = op.abspath(opts.bids_dir)
    subject_list = bids.collect_participants(
        bids_dir, participant_label=opts.participant_label)

    output_dir = op.abspath(opts.output_dir)
    os.makedirs(output_dir, exist_ok=True)

    # Build main workflow
    logger.log(25, INIT_MSG(
        version=info.__version__,
        subject_list=subject_list)
    )

    model = default_path(opts.model, bids_dir, 'model.json')
    preproc_dir = default_path(opts.preproc_dir, output_dir, 'fmriprep')
    deriv_dir = op.join(output_dir, 'fitlins')

    # BIDS-Apps prefers 'participant', BIDS-Model prefers 'subject'
    level = 'subject' if opts.analysis_level == 'participant' else opts.analysis_level

    try:
        retcode = run_model(model, opts.space, level, bids_dir, preproc_dir,
                            deriv_dir)
    except Exception:
        retcode = 1

    sys.exit(retcode)


def run_model(model, space, target_level, bids_dir, preproc_dir, deriv_dir):
    run_context = {'version': info.__version__,
                   'command': ' '.join(sys.argv),
                   'timestamp': time.strftime('%Y-%m-%d %H:%M:%S %z'),
                   }

    analysis = init(model, bids_dir, preproc_dir)
    report_dicts = first_level(analysis, analysis.blocks[0], space, deriv_dir)
    write_report(analysis.blocks[0].level, report_dicts, run_context,
                 deriv_dir)
    if analysis.blocks[0].level == target_level:
        return 0
    for block in analysis.blocks[1:]:
        report_dicts = second_level(analysis, block, space, deriv_dir)
        write_report(block.level, report_dicts, run_context, deriv_dir)
        if block.level == target_level:
            break

    return 0


if __name__ == '__main__':
    raise RuntimeError("fitlins/cli/run.py should not be run directly;\n"
                       "Please `pip install` fitlins and use the `fitlins` command")
